/*
 *  $Id: serverhunter.c,v 1.17 2002/08/22 09:37:25 dreibh Exp $
 *
 * RSerPool implementation.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * Acknowledgement
 * This work was partially funded by the Bundesministerium fr Bildung und
 * Forschung (BMBF) of the Federal Republic of Germany (Frderkennzeichen 01AK045).
 * The authors alone are responsible for the contents.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * There are two mailinglists available at http://www.sctp.de/rserpool.html
 * which should be used for any discussion related to this implementation.
 *
 * Contact: rsplib-discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Name Server Hunter
 *
 */


#include "tdtypes.h"
#include "serverhunter.h"
#include "utilities.h"
#include "netutilities.h"
#include "asapcreator.h"
#include "asapparser.h"

#include <ext_socket.h>
#include <sys/time.h>



#define MAX_SIMULTANEOUS_REQUESTS 3


struct ServerHuntConnection
{
   struct ServerInfo* ServerInfoPtr;
   int                Socket;
   bool               RequestSent;
};



/* ###### Constructor #################################################### */
struct ServerHunter* serverHunterNew()
{
   struct ServerHunter* serverHunter = (struct ServerHunter*)malloc(sizeof(struct ServerHunter));
   if(serverHunter != NULL) {
      serverHunter->ValidServers         = 0;
      serverHunter->ServerList           = serverInfoListNew();
      serverHunter->IPv4ServerHuntSocket = ext_socket(AF_INET,  SOCK_DGRAM, IPPROTO_UDP);
#ifdef HAVE_IPV6
      serverHunter->IPv6ServerHuntSocket = ext_socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
#endif
      serverHunter->Message              = asapMessageNew((char*)&serverHunter->Buffer,sizeof(serverHunter->Buffer));
      if((serverHunter->ServerList == NULL) ||
         (serverHunter->Message    == NULL)) {
         serverHunterDelete(serverHunter);
         serverHunter = NULL;
      }
   }
   return(serverHunter);
}


/* ###### Destructor ##################################################### */
void serverHunterDelete(struct ServerHunter* serverHunter)
{
   if(serverHunter != NULL) {
      if(serverHunter->ServerList) {
         serverInfoListDelete(serverHunter->ServerList);
         serverHunter->ServerList = NULL;
      }
      if(serverHunter->Message) {
         asapMessageDelete(serverHunter->Message);
         serverHunter->Message = NULL;
      }
      if(serverHunter->IPv4ServerHuntSocket >= 0) {
         ext_close(serverHunter->IPv4ServerHuntSocket);
         serverHunter->IPv4ServerHuntSocket = -1;
      }
#ifdef HAVE_IPV6
      if(serverHunter->IPv6ServerHuntSocket >= 0) {
         ext_close(serverHunter->IPv6ServerHuntSocket);
         serverHunter->IPv6ServerHuntSocket = -1;
      }
#endif
      free(serverHunter);
   }
}


/* ###### Send ServerHunt via multicast ################################## */
static void serverHunterServerHuntMulticast(struct ServerHunter* serverHunter,
                                            struct ServerInfo*   serverInfo)
{
   size_t  packetSize;
   ssize_t result;

   LOG_VERBOSE
   fputs("Sending ServerHunt via multicast to ",stdlog);
   transportAddressPrint(serverInfo->Address,stdlog);
   fputs("...\n",stdlog);
   LOG_END

   serverHunter->Message->Type = AHT_SERVER_HUNT;

   packetSize = asapMessage2Packet(serverHunter->Message);
   if(packetSize == 0) {
      LOG_ERROR
      fputs("ServerHunt message creation failed!\n",stdlog);
      LOG_END
      return;
   }

   switch(serverInfo->Address->AddressArray[0].sa.sa_family) {
      case AF_INET:
         LOG_VERBOSE3
         fprintf(stdlog,"Sending via IPv4 multicast socket %d\n",serverHunter->IPv4ServerHuntSocket);
         LOG_END
         result = ext_sendto(serverHunter->IPv4ServerHuntSocket,
                             (char*)&serverHunter->Buffer, packetSize, 0,
                             (struct sockaddr*)&serverInfo->Address->AddressArray[0],
                                getSocklen((struct sockaddr*)&serverInfo->Address->AddressArray[0]));
         if(result < 0) {
            LOG_ERROR
            logerror("Sending ServerHunt via IPv4 multicast failed");
            LOG_END
         }
       break;
#ifdef HAVE_IPV6
      case AF_INET6:
         LOG_VERBOSE3
         fprintf(stdlog,"Sending via IPv6 multicast socket %d\n",serverHunter->IPv6ServerHuntSocket);
         LOG_END
         result = ext_sendto(serverHunter->IPv6ServerHuntSocket,
                             (char*)&serverHunter->Buffer, packetSize, 0,
                             (struct sockaddr*)&serverInfo->Address->AddressArray[0],
                                getSocklen((struct sockaddr*)&serverInfo->Address->AddressArray[0]));
         if(result < 0) {
            LOG_ERROR
            logerror("Sending ServerHunt via IPv6 multicast failed");
            LOG_END
         }
       break;
#endif
      default:
         LOG_WARNING
         fputs("Server address type is unsupported!\n",stdlog);
         LOG_END
       break;
   }
}


/* ###### Connect to server ############################################## */
static bool serverHunterServerHuntUnicastConnect(struct ServerHunter*         serverHunter,
                                                 struct ServerHuntConnection* hunt)
{
   int status;

   hunt->RequestSent = false;
   hunt->Socket      = ext_socket(hunt->ServerInfoPtr->Address->AddressArray[0].sa.sa_family,
                                  SOCK_STREAM,
                                  hunt->ServerInfoPtr->Address->Protocol);

   LOG_VERBOSE
   fputs("Connecting ",stdlog);
   if(hunt->Socket >= 0) {
      fprintf(stdlog,"socket %d ",hunt->Socket);
   }
   fputs("for ServerHunt via unicast to ",stdlog);
   transportAddressPrint(hunt->ServerInfoPtr->Address,stdlog);
   fputs("...\n",stdlog);
   LOG_END

   if(hunt->Socket >= 0) {
      setNonBlocking(hunt->Socket);

      status = ext_connect(hunt->Socket,
                           (struct sockaddr*)&hunt->ServerInfoPtr->Address->AddressArray[0],
                           getSocklen((struct sockaddr*)&hunt->ServerInfoPtr->Address->AddressArray[0]));
      if( ((status < 0) && (errno == EINPROGRESS)) ||
          (status == 0) ) {
         return(true);
      }

      ext_close(hunt->Socket);
      hunt->Socket = -1;
   }

   LOG_VERBOSE
   logerror("Connection establishment for ServerHunt failed");
   LOG_END

   hunt->ServerInfoPtr->Flags |= SF_FAILED;
   LOG_ACTION
   fputs("Invalidated ServerInfo due to connection failure:\n",stdlog);
   serverInfoPrint(hunt->ServerInfoPtr,stdlog);
   LOG_END

   return(false);
}


/* ###### Send ServerHunt via unicast #################################### */
static bool serverHunterServerHuntUnicastWrite(struct ServerHunter*         serverHunter,
                                               struct ServerHuntConnection* hunt)
{
   LOG_VERBOSE
   fputs("Sending ServerHunt via unicast to ",stdlog);
   transportAddressPrint(hunt->ServerInfoPtr->Address,stdlog);
   fprintf(stdlog,", socket %d...\n",hunt->Socket);
   LOG_END

   serverHunter->Message->Type = AHT_SERVER_HUNT;
   if(asapMessageSend(hunt->Socket,0,serverHunter->Message) == true) {
      return(true);
   }

   LOG_VERBOSE
   logerror("Sending ServerHunt failed");
   LOG_END

   hunt->ServerInfoPtr->Flags |= SF_FAILED;
   LOG_ACTION
   fputs("Invalidated ServerInfo due to send failure:\n",stdlog);
   serverInfoPrint(hunt->ServerInfoPtr,stdlog);
   LOG_END

   return(false);
}


/* ###### Handle server response ######################################### */
static bool serverHunterServerHuntHandleResponse(struct ServerHunter*         serverHunter,
                                                 int                          sd,
                                                 struct ServerHuntConnection* hunt)
{
   struct ASAPMessage*      message;
   struct TransportAddress* transportAddress;
   struct sockaddr_storage  senderAddress;
   struct ServerInfo*       serverInfo = NULL;
   socklen_t                senderAddressLength;
   char                     buffer[1024];
   ssize_t                  received;
   bool                     result;

   LOG_VERBOSE
   fprintf(stdlog,"Reading response from socket %d...\n",sd);
   LOG_END

   result              = false;
   senderAddressLength = sizeof(senderAddress);
   received = ext_recvfrom(sd,
                           (char*)&buffer ,sizeof(buffer), 0,
                           (struct sockaddr*)&senderAddress,
                           &senderAddressLength);
   if(received > 0) {
      message = asapPacket2Message((char*)&buffer,received,sizeof(buffer));
      if(message != NULL) {
         if(message->Type == AHT_SERVER_HUNT_RESPONSE) {
            if(message->Error == 0) {

               /* For TCP socket: Try to get peer address if not specified. */
               if(senderAddressLength == 0) {
                  senderAddressLength = sizeof(senderAddress);
                  ext_getpeername(sd,(struct sockaddr*)&senderAddress,
                                 &senderAddressLength);
               }


               LOG_VERBOSE
               fputs("ServerHuntResponse from ",stdlog);
               address2string((struct sockaddr*)&senderAddress,
                              (char*)&buffer, sizeof(buffer), true);
               fputs(buffer,stdlog);
               if(hunt == NULL) {
                  fputs(" (as reply to multicast)",stdlog);
               }
               else {
                  fputs(" (answering unicast request to ",stdlog);
                  transportAddressPrint(hunt->ServerInfoPtr->Address,stdlog);
                  fputs(")",stdlog);
               }
               fputs(" received\n",stdlog);
               LOG_END

               if(hunt == NULL) {
                  if(message->ServerInfoPtr) {
                     message->ServerInfoPtrAutoDelete = false;
                     serverInfo                       = message->ServerInfoPtr;
                     LOG_ACTION
                     fputs("New ServerInfo (from message) for list:\n",stdlog);
                     serverInfoPrint(serverInfo,stdlog);
                     LOG_END
                  }
                  else {
                     transportAddress = transportAddressNew(NAMESERVER_WELL_KNOWN_PROTOCOL,
                                                            NAMESERVER_WELL_KNOWN_PORT,
                                                            (struct sockaddr_storage*)&senderAddress,
                                                            1);
                     if(transportAddress != NULL) {
                        serverInfo = serverInfoNew(0,transportAddress,SF_UNKNOWN);
                        if(serverInfo != NULL) {
                           LOG_ACTION
                           fputs("New ServerInfo (using standard settings) for list:\n",stdlog);
                           serverInfoPrint(serverInfo,stdlog);
                           LOG_END
                        }
                        else {
                           transportAddressDelete(transportAddress);
                        }
                     }
                  }
                  if(serverInfoListAddServerInfo(serverHunter->ServerList,serverInfo) == false) {
                     LOG_ACTION
                     fputs("ServerInfo is already in list\n",stdlog);
                     LOG_END
                     serverInfoDelete(serverInfo);
                  }
               }

               result = true;
            }
         }
         else {
            LOG_WARNING
            fprintf(stdlog,"Unsupported message type $%02x!\n",message->Type);
            LOG_END
         }
         asapMessageDelete(message);
      }
   }

   if(hunt) {
      if(result == true) {
         hunt->ServerInfoPtr->Flags &= ~(SF_UNKNOWN | SF_FAILED);
         LOG_ACTION
         fputs("Validated ServerInfo:\n",stdlog);
         serverInfoPrint(hunt->ServerInfoPtr,stdlog);
         LOG_END
      }
      else {
         hunt->ServerInfoPtr->Flags |= SF_FAILED;
         LOG_ACTION
         fputs("Invalidated ServerInfo due to connection failure:\n",stdlog);
         serverInfoPrint(hunt->ServerInfoPtr,stdlog);
         LOG_END
      }
   }

   return(result);
}


/* ###### Connect to next block of unicast addresses ##################### */
static void serverHunterConnectNextUnicastRequestBlock(struct ServerHunter*         serverHunter,
                                                       struct ServerHuntConnection* connectionSet,
                                                       size_t*                      connections,
                                                       GList**                      list)
{
   struct ServerInfo* serverInfo;

   *connections = 0;
   while(*list != NULL) {
      serverInfo = (struct ServerInfo*)(*list)->data;

      connectionSet[*connections].ServerInfoPtr = serverInfo;
      if(serverHunterServerHuntUnicastConnect(serverHunter,&connectionSet[*connections])) {
         (*connections)++;
      }
      if(*connections == MAX_SIMULTANEOUS_REQUESTS) {
         break;
      }

      *list = g_list_next(*list);
   }
}


/* ###### Send ServerHunt via all multicast addresses #################### */
static void serverHunterSendMulticastRequestBlock(struct ServerHunter* serverHunter)
{
   struct ServerInfo* serverInfo;
   GList* list = g_list_first(serverHunter->ServerList->MulticastServerList);
   while(list != NULL) {
      serverInfo = (struct ServerInfo*)list->data;
      serverHunterServerHuntMulticast(serverHunter,serverInfo);
      list = g_list_next(list);
   }
}


/* ###### Do server hunt ################################################# */
unsigned int serverHunterRun(struct ServerHunter* serverHunter,
                             int*                 sd,
                             const card64         serverHuntTimeout,
                             const card64         maxServerHuntTimeout,
                             const unsigned int   maxServerHuntTrials,
                             const char*          configFile)
{
   struct ServerHuntConnection connectionSet[MAX_SIMULTANEOUS_REQUESTS];
   struct timeval     selectTimeout;
   GList*             list;
   fd_set             wfdset;
   fd_set             rfdset;
   fd_set             efdset;
   size_t             connections;
   int                n;
   int                i;
   int                result;
   unsigned int       outstanding;
   card64             now;
   card64             start;
   card64             remaining;
   card64             timeout;
   unsigned int       trials;


   if(serverHunter == NULL) {
      return(0);
   }
   LOG_ACTION
   fputs("Starting Server Hunt...\n",stdlog);
   LOG_END


   serverHunter->ValidServers = 0;
   for(i = 0;i < MAX_SIMULTANEOUS_REQUESTS;i++) {
      connectionSet[i].Socket        = -1;
      connectionSet[i].ServerInfoPtr = NULL;
      connectionSet[i].RequestSent   = false;
   }
   if(sd != NULL) {
      *sd = -1;
   }


   timeout = serverHuntTimeout;
   trials  = 0;
   start   = getMicroTime();

   LOG_ACTION
   fprintf(stdlog,"Server Hunt Trial #1 with timeout %Ld [s]\n",timeout);
   LOG_END


   serverInfoListGetStaticEntries(serverHunter->ServerList,configFile);

   serverHunterSendMulticastRequestBlock(serverHunter);

   list = g_list_first(serverHunter->ServerList->UnicastServerList);
   serverHunterConnectNextUnicastRequestBlock(serverHunter,
                                              (struct ServerHuntConnection*)&connectionSet,
                                              &connections,
                                              &list);

   for(;;) {
      n = 0;
      FD_ZERO(&rfdset);
      FD_ZERO(&wfdset);
      FD_ZERO(&efdset);
      if(serverHunter->IPv4ServerHuntSocket >= 0) {
         FD_SET(serverHunter->IPv4ServerHuntSocket,&rfdset);
         FD_SET(serverHunter->IPv4ServerHuntSocket,&efdset);
         n = max(n,serverHunter->IPv4ServerHuntSocket);
      }
#ifdef HAVE_IPV6
      if(serverHunter->IPv6ServerHuntSocket >= 0) {
         FD_SET(serverHunter->IPv6ServerHuntSocket,&rfdset);
         FD_SET(serverHunter->IPv6ServerHuntSocket,&efdset);
         n = max(n,serverHunter->IPv6ServerHuntSocket);
      }
#endif
      outstanding = 0;
      for(i = 0;i < MAX_SIMULTANEOUS_REQUESTS;i++) {
         if(connectionSet[i].Socket >= 0) {
            FD_SET(connectionSet[i].Socket,&rfdset);
            FD_SET(connectionSet[i].Socket,&efdset);
            if(connectionSet[i].RequestSent == false) {
               FD_SET(connectionSet[i].Socket,&wfdset);
            }
            n = max(n,connectionSet[i].Socket);
            outstanding++;
         }
      }

      now = getMicroTime();
      if((now - start) < timeout) {
         remaining = timeout - (now - start);
      }
      else {
         remaining = 0;
      }

      selectTimeout.tv_sec  = remaining / (card64)1000000;
      selectTimeout.tv_usec = remaining % (card64)1000000;
      LOG_VERBOSE3
      fprintf(stdlog,"select() with timeout %Ld\n",remaining);
      LOG_END
      result = ext_select(n + 1,
                          &rfdset, &wfdset, &efdset,
                          &selectTimeout);
      LOG_VERBOSE3
      fprintf(stdlog,"select() result=%d\n",result);
      LOG_END
      if(result == 0) {
         serverHunterSendMulticastRequestBlock(serverHunter);
         serverHunterConnectNextUnicastRequestBlock(serverHunter,
                                                    (struct ServerHuntConnection*)&connectionSet,
                                                    &connections,
                                                    &list);
         if(list == NULL) {
            trials++;
            if(trials >= maxServerHuntTrials) {
               if(serverHunter->ValidServers == 0) {
                  LOG_ERROR
                  fputs("Server Hunt failed, no usable servers found\n",stdlog);
                  LOG_END
               }
               break;
            }

            start = getMicroTime();
            list = g_list_first(serverHunter->ServerList->UnicastServerList);
            timeout *= 2;
            if(timeout > maxServerHuntTimeout) {
               timeout = maxServerHuntTimeout;
            }
            LOG_ACTION
            fprintf(stdlog,"Server Hunt Trial #%d with timeout %Ld [s]\n",trials + 1,timeout);
            LOG_END
         }
      }
      else if(result > 0) {
         if(serverHunter->IPv4ServerHuntSocket >= 0) {
            if(FD_ISSET(serverHunter->IPv4ServerHuntSocket,&rfdset)) {
               LOG_VERBOSE3
               fprintf(stdlog,"Read event for IPv4 multicast socket %d\n",serverHunter->IPv4ServerHuntSocket);
               LOG_END
               if(serverHunterServerHuntHandleResponse(serverHunter,
                                                       serverHunter->IPv4ServerHuntSocket,
                                                       NULL)) {
                  if(list == NULL) {
                     list = g_list_last(serverHunter->ServerList->UnicastServerList);
                  }
               }
            }
         }
#ifdef HAVE_IPV6
         if(serverHunter->IPv6ServerHuntSocket >= 0) {
            if(FD_ISSET(serverHunter->IPv6ServerHuntSocket,&rfdset)) {
               LOG_VERBOSE3
               fprintf(stdlog,"Read event for IPv6 multicast socket %d\n",serverHunter->IPv6ServerHuntSocket);
               LOG_END
               if(serverHunterServerHuntHandleResponse(serverHunter,
                                                       serverHunter->IPv6ServerHuntSocket,
                                                       NULL)) {
                  if(list == NULL) {
                     list = g_list_last(serverHunter->ServerList->UnicastServerList);
                  }
               }
            }
         }
#endif

         for(i = 0;i < MAX_SIMULTANEOUS_REQUESTS;i++) {
            if(connectionSet[i].Socket >= 0) {
               if(FD_ISSET(connectionSet[i].Socket,&efdset)) {
                  LOG_ACTION
                  fprintf(stdlog,"Exception on socket %d\n",connectionSet[i].Socket);
                  LOG_END
                  ext_close(connectionSet[i].Socket);
                  connectionSet[i].Socket = -1;
                  connections--;
               }
            }
            if(connectionSet[i].Socket >= 0) {
               if(FD_ISSET(connectionSet[i].Socket,&wfdset)) {
                  if(serverHunterServerHuntUnicastWrite(serverHunter,
                                                        &connectionSet[i]) == false) {
                     ext_close(connectionSet[i].Socket);
                     connectionSet[i].Socket = -1;
                     connections--;
                  }
                  connectionSet[i].RequestSent = true;
               }
            }
            if(connectionSet[i].Socket >= 0) {
               if(FD_ISSET(connectionSet[i].Socket,&rfdset)) {
                  LOG_VERBOSE3
                  fprintf(stdlog,"Read event for unicast socket %d\n",connectionSet[i].Socket);
                  LOG_END
                  if(serverHunterServerHuntHandleResponse(serverHunter,
                                                          connectionSet[i].Socket,
                                                          &connectionSet[i])) {
                     serverHunter->ValidServers++;

                     LOG_ACTION
                     fprintf(stdlog,"Valid server found on socket %d\n",connectionSet[i].Socket);
                     serverInfoPrint(connectionSet[i].ServerInfoPtr,stdlog);
                     LOG_END

                     if(sd != NULL) {
                        *sd = connectionSet[i].Socket;
                        connectionSet[i].Socket = -1;
                        goto complete;
                     }
                  }
                  if(connectionSet[i].Socket >= 0) {
                     ext_close(connectionSet[i].Socket);
                     connectionSet[i].Socket = -1;
                  }
                  break;
               }
            }
         }

         if(connections == 0) {
            if(list != NULL) {
               LOG_ACTION
               fputs("Trying next block of unicast addresses...\n",stdlog);
               LOG_END
               serverHunterConnectNextUnicastRequestBlock(serverHunter,
                                                          (struct ServerHuntConnection*)&connectionSet,
                                                          &connections,
                                                          &list);
            }
         }
      }
      else {
         LOG_ERROR
         logerror("select() call failed");
         LOG_END
         break;
      }
   }


complete:
   for(i = 0;i < MAX_SIMULTANEOUS_REQUESTS;i++) {
      ext_close(connectionSet[i].Socket);
      connectionSet[i].Socket = -1;
   }
   serverInfoListPurge(serverHunter->ServerList,(card64)-1);

   return(serverHunter->ValidServers);
}
